@@ -5,6 +5,7 @@ import os
 import pathlib
 import time
 import warnings
+import weakref
 
 from packaging.version import Version
 
@@ -12,13 +13,13 @@ from . import _compression as mcompression
 from . import _display as display
 from . import _node_info as node_info
 from . import _version as version
-from . import constants, generic_io, reference, schema, treeutil, util, versioning, yamlutil
+from . import constants, generic_io, lazy_nodes, reference, schema, treeutil, util, versioning, yamlutil
 from ._block.manager import Manager as BlockManager
 from ._helpers import validate_version
 from .config import config_context, get_config
 from .exceptions import (
-    AsdfConversionWarning,
-    AsdfDeprecationWarning,
+    AsdfManifestURIMismatchWarning,
+    AsdfPackageVersionWarning,
     AsdfWarning,
     DelimiterNotFoundError,
     ValidationError,
@@ -29,21 +30,6 @@ from .tags.core import AsdfObject, ExtensionMetadata, HistoryEntry, Software
 from .util import NotSet
 
 
-def __getattr__(name):
-    if name == "SerializationContext":
-        warnings.warn(
-            "importing SerializationContext from asdf.asdf is deprecated. "
-            "Please import SerializationContext from asdf.extension",
-            AsdfDeprecationWarning,
-        )
-        from .extension._serialization_context import SerializationContext
-
-        return SerializationContext
-
-    msg = f"module {__name__!r} has no attribute {name!r}"
-    raise AttributeError(msg)
-
-
 def _get_asdf_library_info():
     """
     Get information about asdf to include in the asdf_library entry
@@ -70,11 +56,8 @@ class AsdfFile:
         uri=None,
         extensions=None,
         version=None,
-        ignore_version_mismatch=True,
         ignore_unrecognized_tag=False,
-        ignore_implicit_conversion=False,
-        copy_arrays=False,
-        memmap=NotSet,
+        memmap=False,
         lazy_load=True,
         custom_schema=None,
     ):
@@ -99,31 +82,13 @@ class AsdfFile:
             The ASDF Standard version.  If not provided, defaults to the
             configured default version.  See `asdf.config.AsdfConfig.default_version`.
 
-        ignore_version_mismatch : bool, optional
-            When `True`, do not raise warnings for mismatched schema versions.
-            Set to `True` by default.
-
         ignore_unrecognized_tag : bool, optional
             When `True`, do not raise warnings for unrecognized tags. Set to
             `False` by default.
 
-        ignore_implicit_conversion : bool
-            When `True`, do not raise warnings when types in the tree are
-            implicitly converted into a serializable object. The motivating
-            case for this is currently ``namedtuple``, which cannot be serialized
-            as-is.
-
-        copy_arrays : bool, optional
-            When `False`, when reading files, attempt to memmap underlying data
-            arrays when possible.
-
         memmap : bool, optional
             When `True`, when reading files, attempt to memmap underlying data
-            arrays when possible. When set, this argument will override
-            ``copy_arrays``.  When not set, the ``copy_arrays`` will determine
-            if arrays are memory mapped or copied. ``copy_arrays`` will be
-            deprecated and the default will change in an upcoming asdf version
-            which by default will not memory map arrays.
+            arrays when possible. Defaults to ``False``.
 
         lazy_load : bool, optional
             When `True` and the underlying file handle is seekable, data
@@ -132,8 +97,6 @@ class AsdfFile:
             open during the lifetime of the tree. Setting to False causes
             all data arrays to be loaded up front, which means that they
             can be accessed even after the underlying file is closed.
-            Note: even if ``lazy_load`` is `False`, ``copy_arrays`` is still taken
-            into account.
 
         custom_schema : str, optional
             Path to a custom schema file that will be used for a secondary
@@ -162,32 +125,21 @@ class AsdfFile:
         else:
             self._custom_schema = None
 
-        self._ignore_version_mismatch = ignore_version_mismatch
         self._ignore_unrecognized_tag = ignore_unrecognized_tag
-        self._ignore_implicit_conversion = ignore_implicit_conversion
-
-        # Set of (string, string) tuples representing tag version mismatches
-        # that we've already warned about for this file.
-        self._warned_tag_pairs = set()
 
         # Context of a call to treeutil.walk_and_modify, needed in the AsdfFile
         # in case walk_and_modify is re-entered by extension code (via
         # custom_tree_to_tagged_tree or tagged_tree_to_custom_tree).
         self._tree_modification_context = treeutil._TreeModificationContext()
 
+        # A cache of tagged objects and their converted custom objects used when
+        # a file is read with "lazy_tree=True". Used by lazy_nodes.
+        self._tagged_object_cache = lazy_nodes._TaggedObjectCache()
+
         self._fd = None
         self._closed = False
         self._external_asdf_by_uri = {}
-        # if memmap is set, it overrides copy_arrays
-        if memmap is not NotSet:
-            copy_arrays = not memmap
-        self._blocks = BlockManager(uri=uri, lazy_load=lazy_load, memmap=not copy_arrays)
-        # this message is passed into find_references to only warn if
-        # a reference was found
-        find_ref_warning_msg = (
-            "find_references during AsdfFile.__init__ is deprecated. "
-            "call AsdfFile.find_references after AsdfFile.__init__"
-        )
+        self._blocks = BlockManager(uri=uri, lazy_load=lazy_load, memmap=memmap)
         if tree is None:
             # Bypassing the tree property here, to avoid validating
             # an empty tree.
@@ -201,21 +153,12 @@ class AsdfFile:
             self._blocks._uri = tree.uri
             # Set directly to self._tree (bypassing property), since
             # we can assume the other AsdfFile is already valid.
-            self._tree = tree.tree
-            self.find_references(_warning_msg=find_ref_warning_msg)
+            # Call "walk_and_modify" here to use it's logic for
+            # creating a "copy" of the other tree. This mimics what was previously
+            # a call to find_references (which we longer do in AsdfFile.__init__)
+            self._tree = treeutil.walk_and_modify(tree.tree, lambda o: o)
         else:
             self._tree = AsdfObject(tree)
-            try:
-                self.validate()
-            except ValidationError:
-                warnings.warn(
-                    "Validation during AsdfFile.__init__ is deprecated. "
-                    "Please use AsdfFile.validate to validate the tree",
-                    AsdfDeprecationWarning,
-                )
-                raise
-
-            self.find_references(_warning_msg=find_ref_warning_msg)
 
         self._comments = []
 
@@ -257,14 +200,6 @@ class AsdfFile:
         """
         return str(self._version)
 
-    @property
-    def version_map(self):
-        warnings.warn(
-            "AsdfFile.version_map is deprecated. Please use the extension_manager",
-            AsdfDeprecationWarning,
-        )
-        return versioning._get_version_map(self.version_string)
-
     @property
     def extensions(self):
         """
@@ -356,7 +291,7 @@ class AsdfFile:
                 if strict:
                     raise RuntimeError(msg)
 
-                warnings.warn(msg, AsdfWarning)
+                warnings.warn(msg, AsdfPackageVersionWarning)
 
             elif extension.software:
                 # Local extensions may not have a real version.  If the package name changed,
@@ -372,7 +307,33 @@ class AsdfFile:
                     if strict:
                         raise RuntimeError(msg)
 
-                    warnings.warn(msg, AsdfWarning)
+                    warnings.warn(msg, AsdfPackageVersionWarning)
+
+            # check version of manifest providing package (if one was recorded)
+            if "manifest_software" in extension:
+                package_name = extension["manifest_software"]["name"]
+                package_version = Version(extension["manifest_software"]["version"])
+                package_description = f"{package_name}=={package_version}"
+                installed_version = None
+                for mapping in get_config().resource_manager._resource_mappings:
+                    if mapping.package_name == package_name:
+                        installed_version = Version(mapping.package_version)
+                        break
+                msg = None
+                if installed_version is None:
+                    msg = (
+                        f"File {filename}was created with package {package_description}, "
+                        "which is currently not installed"
+                    )
+                elif installed_version < package_version:
+                    msg = (
+                        f"File {filename}was created with package {package_description}, "
+                        f"but older package({package_name}=={installed_version}) is installed."
+                    )
+                if msg:
+                    if strict:
+                        raise RuntimeError(msg)
+                    warnings.warn(msg, AsdfPackageVersionWarning)
 
     def _process_plugin_extensions(self):
         """
@@ -460,6 +421,23 @@ class AsdfFile:
                 ext_meta["extension_uri"] = extension.extension_uri
             if extension.compressors:
                 ext_meta["supported_compression"] = [comp.label.decode("ascii") for comp in extension.compressors]
+            manifest = getattr(extension._delegate, "_manifest", None)
+            if manifest is not None:
+                # check if this extension was built from a manifest is a different package
+                resource_mapping = get_config().resource_manager._mappings_by_uri.get(manifest["id"])
+                # if an extension was registered with a manifest uri that does not match the id the
+                # resource_mapping lookup will fail (resource_mapping will be None)
+                if resource_mapping is None:
+                    warnings.warn(
+                        f"Extension ({extension.extension_uri}) uses a manifest with a uri that "
+                        "does not match it's id. Please open an issue with the extension maintainer",
+                        AsdfManifestURIMismatchWarning,
+                    )
+                elif resource_mapping.package_name != extension.package_name:
+                    ext_meta["manifest_software"] = Software(
+                        name=resource_mapping.package_name,
+                        version=resource_mapping.package_version,
+                    )
 
             for i, entry in enumerate(tree["history"]["extensions"]):
                 # Update metadata about this extension if it already exists
@@ -490,6 +468,7 @@ class AsdfFile:
             # as we're closing the file, also empty out the
             # tree so that references to array data can be released
             self._tree = AsdfObject()
+            self._tagged_object_cache.clear()
         for external in self._external_asdf_by_uri.values():
             external.close()
         self._external_asdf_by_uri.clear()
@@ -580,16 +559,7 @@ class AsdfFile:
 
     @tree.setter
     def tree(self, tree):
-        asdf_object = AsdfObject(tree)
-        # Only perform custom validation if the tree is not empty
-        try:
-            self._validate(asdf_object, custom=bool(tree))
-        except ValidationError:
-            warnings.warn(
-                "Validation on tree assignment is deprecated. Please use AsdfFile.validate", AsdfDeprecationWarning
-            )
-            raise
-        self._tree = asdf_object
+        self._tree = AsdfObject(tree)
 
     def keys(self):
         return self.tree.keys()
@@ -734,6 +704,41 @@ class AsdfFile:
         """ """
         return self._blocks._get_array_compression_kwargs(arr)
 
+    def set_array_save_base(self, arr, save_base):
+        """
+        Set the ``save_base`` option for ``arr``. When ``arr`` is
+        written to a file, if ``save_base`` is ``True`` the base array
+        for ``arr`` will be saved.
+
+        Note that similar to other array options this setting is linked
+        to the base array if ``arr`` is a view.
+
+        Parameters
+        ----------
+        arr : numpy.ndarray
+
+        save_base : bool or None
+            if ``None`` the ``default_array_save_base`` value from asdf
+            config will be used
+        """
+        self._blocks._set_array_save_base(arr, save_base)
+
+    def get_array_save_base(self, arr):
+        """
+        Returns the ``save_base`` option for ``arr``. When ``arr`` is
+        written to a file, if ``save_base`` is ``True`` the base array
+        for ``arr`` will be saved.
+
+        Parameters
+        ----------
+        arr : numpy.ndarray
+
+        Returns
+        -------
+        save_base : bool or None
+        """
+        return self._blocks._get_array_save_base(arr)
+
     @classmethod
     def _parse_header_line(cls, line):
         """
@@ -802,6 +807,7 @@ class AsdfFile:
         fd,
         validate_checksums=False,
         extensions=None,
+        lazy_tree=NotSet,
         _get_yaml_content=False,
         _force_raw_types=False,
         strict_extension_check=False,
@@ -813,7 +819,7 @@ class AsdfFile:
             msg = "'strict_extension_check' and 'ignore_missing_extensions' are incompatible options"
             raise ValueError(msg)
 
-        with config_context():
+        with config_context() as cfg:
             # validate_checksums (unlike memmap and lazy_load) is provided
             # here instead of in __init__
             self._blocks._validate_checksums = validate_checksums
@@ -881,9 +887,6 @@ class AsdfFile:
                 # to select the correct tag for us.
                 tree = yamlutil.custom_tree_to_tagged_tree(AsdfObject(), self)
 
-            find_ref_warning_msg = "find_references during open is deprecated. call AsdfFile.find_references after open"
-            tree = reference.find_references(tree, self, _warning_msg=find_ref_warning_msg)
-
             if self.version <= versioning.FILL_DEFAULTS_MAX_VERSION and get_config().legacy_fill_schema_defaults:
                 schema.fill_defaults(tree, self, reading=True)
 
@@ -894,7 +897,14 @@ class AsdfFile:
                     self.close()
                     raise
 
-            tree = yamlutil.tagged_tree_to_custom_tree(tree, self, _force_raw_types)
+            if lazy_tree is NotSet:
+                lazy_tree = cfg.lazy_tree
+            if lazy_tree and not _force_raw_types:
+                obj = AsdfObject()
+                obj.data = lazy_nodes.AsdfDictNode(tree, weakref.ref(self))
+                tree = obj
+            else:
+                tree = yamlutil.tagged_tree_to_custom_tree(tree, self, _force_raw_types)
 
             if not (ignore_missing_extensions or _force_raw_types):
                 self._check_extensions(tree, strict=strict_extension_check)
@@ -912,6 +922,7 @@ class AsdfFile:
         mode="r",
         validate_checksums=False,
         extensions=None,
+        lazy_tree=NotSet,
         _get_yaml_content=False,
         _force_raw_types=False,
         strict_extension_check=False,
@@ -926,6 +937,7 @@ class AsdfFile:
                 generic_file,
                 validate_checksums=validate_checksums,
                 extensions=extensions,
+                lazy_tree=lazy_tree,
                 _get_yaml_content=_get_yaml_content,
                 _force_raw_types=_force_raw_types,
                 strict_extension_check=strict_extension_check,
@@ -1241,49 +1253,22 @@ class AsdfFile:
                 if version is not None:
                     self.version = previous_version
 
-    def find_references(self, _warning_msg=False):
+    def find_references(self):
         """
         Finds all external "JSON References" in the tree and converts
         them to ``reference.Reference`` objects.
         """
         # Set directly to self._tree, since it doesn't need to be re-validated.
-        self._tree = reference.find_references(self._tree, self, _warning_msg=_warning_msg)
+        self._tree = reference.find_references(self._tree, self)
 
-    def resolve_references(self, **kwargs):
+    def resolve_references(self):
         """
         Finds all external "JSON References" in the tree, loads the
         external content, and places it directly in the tree.  Saving
         a ASDF file after this operation means it will have no
         external references, and will be completely self-contained.
         """
-        if len(kwargs):
-            warnings.warn("Passing kwargs to resolve_references is deprecated and does nothing", AsdfDeprecationWarning)
         self._tree = reference.resolve_references(self._tree, self)
-        try:
-            self.validate()
-        except ValidationError:
-            warnings.warn(
-                "Validation during resolve_references is deprecated. "
-                "Please use AsdfFile.validate after resolve_references to validate the resolved tree",
-                AsdfDeprecationWarning,
-            )
-            raise
-
-    def resolve_and_inline(self):
-        """
-        Resolves all external references and inlines all data.  This
-        produces something that, when saved, is a 100% valid YAML
-        file.
-        """
-        warnings.warn(
-            "resolve_and_inline is deprecated. "
-            "Use AsdfFile.resolve_references and all_array_storage=inline "
-            "during AsdfFile.write_to",
-            AsdfDeprecationWarning,
-        )
-        self.resolve_references()
-        for b in self._blocks.blocks:
-            self.set_array_storage(b.data, "inline")
 
     def fill_defaults(self):
         """
@@ -1511,20 +1496,6 @@ class AsdfFile:
         result = AsdfSearchResult(["root"], self.tree)
         return result.search(key=key, type_=type_, value=value, filter_=filter_)
 
-    # This function is called from within TypeIndex when deserializing
-    # the tree for this file.  It is kept here so that we can keep
-    # state on the AsdfFile and prevent a flood of warnings for the
-    # same tag.
-    def _warn_tag_mismatch(self, tag, best_tag):
-        if not self._ignore_version_mismatch and (tag, best_tag) not in self._warned_tag_pairs:
-            message = (
-                f"No explicit ExtensionType support provided for tag '{tag}'. "
-                f"The ExtensionType subclass for tag '{best_tag}' will be used instead. "
-                "This fallback behavior will be removed in asdf 3.0."
-            )
-            warnings.warn(message, AsdfConversionWarning)
-            self._warned_tag_pairs.add((tag, best_tag))
-
     # This function is called from within yamlutil methods to create
     # a context when one isn't explicitly passed in.
     def _create_serialization_context(self, operation=_serialization_context.BlockAccess.NONE):
@@ -1537,11 +1508,10 @@ def open_asdf(
     mode=None,
     validate_checksums=False,
     extensions=None,
-    ignore_version_mismatch=True,
     ignore_unrecognized_tag=False,
     _force_raw_types=False,
-    copy_arrays=False,
-    memmap=NotSet,
+    memmap=False,
+    lazy_tree=NotSet,
     lazy_load=True,
     custom_schema=None,
     strict_extension_check=False,
@@ -1573,25 +1543,13 @@ def open_asdf(
         Additional extensions to use when reading and writing the file.
         May be an `asdf.extension.Extension` or a `list` of extensions.
 
-    ignore_version_mismatch : bool, optional
-        When `True`, do not raise warnings for mismatched schema versions.
-        Set to `True` by default.
-
     ignore_unrecognized_tag : bool, optional
         When `True`, do not raise warnings for unrecognized tags. Set to
         `False` by default.
 
-    copy_arrays : bool, optional
-        When `False`, when reading files, attempt to memmap underlying data
-        arrays when possible.
-
     memmap : bool, optional
         When `True`, when reading files, attempt to memmap underlying data
-        arrays when possible. When set, this argument will override
-        ``copy_arrays``.  When not set, the ``copy_arrays`` will determine
-        if arrays are memory mapped or copied. ``copy_arrays`` will be
-        deprecated and the default will change in an upcoming asdf version
-        which by default will not memory map arrays.
+        arrays when possible. Defaults to ``False``.
 
     lazy_load : bool, optional
         When `True` and the underlying file handle is seekable, data
@@ -1603,6 +1561,15 @@ def open_asdf(
         Note: even if ``lazy_load`` is `False`, ``memmap`` is still taken
         into account.
 
+    lazy_tree : bool, optional
+        When `True` the ASDF tree will not be converted to custom objects
+        when the file is loaded. Instead, objects will be "lazily" converted
+        only when they are accessed. Note that the tree will not contain dict
+        and list instances for containers and instead return instances of classes
+        defined in `asdf.lazy_nodes`. Since objects are converted when they
+        are accessed, traversing the tree (like is done during `AsdfFile.info`
+        and `AsdfFile.search`) will result in nodes being converted.
+
     custom_schema : str, optional
         Path to a custom schema file that will be used for a secondary
         validation pass. This can be used to ensure that particular ASDF
@@ -1641,9 +1608,7 @@ def open_asdf(
             mode = "r"
 
     instance = AsdfFile(
-        ignore_version_mismatch=ignore_version_mismatch,
         ignore_unrecognized_tag=ignore_unrecognized_tag,
-        copy_arrays=copy_arrays,
         memmap=memmap,
         lazy_load=lazy_load,
         custom_schema=custom_schema,
@@ -1656,6 +1621,7 @@ def open_asdf(
         mode=mode,
         validate_checksums=validate_checksums,
         extensions=extensions,
+        lazy_tree=lazy_tree,
         _get_yaml_content=_get_yaml_content,
         _force_raw_types=_force_raw_types,
         strict_extension_check=strict_extension_check,
